from linkml_transformer.datamodel.transformer_model import TransformationSpecification


def invert_transformation_specification(
    specification: TransformationSpecification,
) -> TransformationSpecification:
    raise NotImplementedError
